/*
 *
 * Copyright 2017 Shuouma <dreamcast-talk.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 * A simple Meta Server for Max Pool for Dreamcast
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define BUFSIZE 1024
#define HDRSIZE 15

/*
 * error - wrapper for perror
 */
void error(char *msg) {
  perror(msg);
  exit(1);
}

int uint16_to_char(uint16_t data, char* msg) {
  msg[0] = (char)(data >> 8);
  msg[1] = (char)data;
  return 2;
}

int uint32_to_char(uint32_t data, char* msg) {
  msg[0] = (char)(data >> 24);
  msg[1] = (char)(data >> 16);
  msg[2] = (char)(data >> 8);
  msg[3] = (char)data;
  return 4;
}

void create_header(char *msg, uint16_t nr_of_servers) {
  msg[0] = 0x05;
  msg[1] = 0x02;
  msg[8] = 0x0A;
  msg[9] = 0x00;
  msg[10] = 0x00;
  msg[11] = 0x04;
  uint16_to_char(htons(nr_of_servers), &msg[12]);
  msg[14] = 0x00;
}

void fetch_ip_from_hostname(char* ip) {
  
  char * hostname = "dctalk.no-ip.info";
  struct hostent *he;
  struct in_addr **addr_list;
  int i;
         
  if ( (he = gethostbyname( hostname ) ) == NULL)  {
    printf("%s is unavailable\n", hostname);
    strcpy(ip, "173.69.21.183");
    return;
  }
 
  addr_list = (struct in_addr **) he->h_addr_list;
     
  for(i = 0; addr_list[i] != NULL; i++) {
    strcpy(ip , inet_ntoa(*addr_list[i]) );
    return;
  }
}

uint16_t create_server_list(char *msg) {
  uint16_t nr_of_servers = 1;
  uint32_t ip = 0;
  uint16_t port = 0;
  struct sockaddr_in sa = {0};
  char ip_str[100];

  fetch_ip_from_hostname(ip_str);
  
  inet_pton(AF_INET, ip_str, &(sa.sin_addr));
  ip = ntohl(sa.sin_addr.s_addr);
  port = htons(35000);

  uint16_to_char(htons(port), &msg[0]);
  uint32_to_char(ip, &msg[2]);
    
  return nr_of_servers;
}

int verify_request(char * msg, ssize_t len) {
  unsigned char verify_array[9] = { 0x05, 0x02, 0x00, 0x66, 0x00, 0x0a, 0x00, 0x00, 0x04 };
  if (len < 9) {
    return -1;
  }
  if (memcmp(msg, verify_array, 9) != 0) {
    return -1;
  }

  return 0;
}

int main(int argc, char **argv) {
  int sockfd; /* socket */
  int portno; /* port to listen on */
  int clientlen; /* byte size of client's address */
  struct sockaddr_in serveraddr; /* server's addr */
  struct sockaddr_in clientaddr; /* client addr */
  char buf[BUFSIZE]; /* message buf */
  char *hostaddrp; /* dotted decimal host addr string */
  int optval; /* flag value for setsockopt */
  ssize_t n; /* message byte size */
  
  portno = 6003;

  sockfd = socket(AF_INET, SOCK_DGRAM, 0);
  if (sockfd < 0) 
    error("ERROR opening socket");
  
  optval = 1;
  setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, 
	     (const void *)&optval , sizeof(int));

  bzero((char *) &serveraddr, sizeof(serveraddr));
  serveraddr.sin_family = AF_INET;
  serveraddr.sin_addr.s_addr = htonl(INADDR_ANY);
  serveraddr.sin_port = htons((unsigned short)portno);

  if (bind(sockfd, (struct sockaddr *) &serveraddr, 
	   sizeof(serveraddr)) < 0) 
    error("ERROR on binding");

  clientlen = sizeof(clientaddr);
  while (1) {

    bzero(buf, BUFSIZE);
    n = recvfrom(sockfd, buf, BUFSIZE, (ssize_t)0,
		 (struct sockaddr *) &clientaddr, (socklen_t*)&clientlen);
    if (n < 0)
      error("ERROR in recvfrom");

    if (verify_request(buf, n) != 0) {
      continue;
    }

    hostaddrp = inet_ntoa(clientaddr.sin_addr);
    if (hostaddrp == NULL)
      error("ERROR on inet_ntoa\n");

    memset(buf, 0, sizeof(buf));
    uint16_t nr_of_servers = create_server_list(&buf[HDRSIZE]);
    create_header(buf, nr_of_servers);
    int pkt_size = (int)(nr_of_servers*6)+HDRSIZE;
    
    n = sendto(sockfd, buf, (size_t)pkt_size, 0, 
	       (struct sockaddr *) &clientaddr, (socklen_t)clientlen);
    if (n < 0) 
      error("ERROR in sendto");
  }
}
